@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
@@ -74,13 +96,8 @@ pdf_set_cmap_wmode(fz_context *ctx, pdf_cmap *cmap, int wmode)
 	cmap->wmode = wmode;
 }
 
-/*
- * Add a codespacerange section.
- * These ranges are used by pdf_decode_cmap to decode
- * multi-byte encoded strings.
- */
 void
-pdf_add_codespace(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned int high, int n)
+pdf_add_codespace(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned int high, size_t n)
 {
 	if (cmap->codespace_len + 1 == nelem(cmap->codespace))
 	{
@@ -88,14 +105,19 @@ pdf_add_codespace(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned in
 		return;
 	}
 
-	cmap->codespace[cmap->codespace_len].n = n;
+	if ((uint32_t)n != n)
+	{
+		fz_warn(ctx, "assert: code space range too large");
+		return;
+	}
+
+	cmap->codespace[cmap->codespace_len].n = (int)n;
 	cmap->codespace[cmap->codespace_len].low = low;
 	cmap->codespace[cmap->codespace_len].high = high;
 	cmap->codespace_len ++;
 }
 
-
-struct cmap_splay_s {
+struct cmap_splay {
 	unsigned int low;
 	unsigned int high;
 	unsigned int out;
@@ -282,7 +304,7 @@ static unsigned int delete_node(pdf_cmap *cmap, unsigned int current)
 	else
 	{
 		/* Hard case, find the in-order predecessor of current */
-		int amputee = current;
+		unsigned int amputee = current;
 		replacement = tree[current].left;
 		while (tree[replacement].right != EMPTY) {
 			amputee = replacement;
@@ -325,29 +347,29 @@ static unsigned int delete_node(pdf_cmap *cmap, unsigned int current)
 
 	/* current is now unlinked. We need to remove it from our array. */
 	cmap->tlen--;
-	if (current != cmap->tlen)
+	if (current != (unsigned int) cmap->tlen)
 	{
-		if (replacement == cmap->tlen)
+		if (replacement == (unsigned int) cmap->tlen)
 			replacement = current;
 		tree[current] = tree[cmap->tlen];
 		parent = tree[current].parent;
 		if (parent == EMPTY)
 			cmap->ttop = current;
-		else if (tree[parent].left == cmap->tlen)
+		else if (tree[parent].left == (unsigned int) cmap->tlen)
 			tree[parent].left = current;
 		else
 		{
-			assert(tree[parent].right == cmap->tlen);
+			assert(tree[parent].right == (unsigned int) cmap->tlen);
 			tree[parent].right = current;
 		}
 		if (tree[current].left != EMPTY)
 		{
-			assert(tree[tree[current].left].parent == cmap->tlen);
+			assert(tree[tree[current].left].parent == (unsigned int) cmap->tlen);
 			tree[tree[current].left].parent = current;
 		}
 		if (tree[current].right != EMPTY)
 		{
-			assert(tree[tree[current].right].parent == cmap->tlen);
+			assert(tree[tree[current].right].parent == (unsigned int) cmap->tlen);
 			tree[tree[current].right].parent = current;
 		}
 	}
@@ -362,7 +384,7 @@ dump_splay(cmap_splay *tree, unsigned int node, int depth, const char *pre)
 {
 	int i;
 
-	if (node == EMPTY)
+	if (tree == NULL || node == EMPTY)
 		return;
 
 	for (i = 0; i < depth; i++)
@@ -440,6 +462,19 @@ static void walk_splay(cmap_splay *tree, unsigned int node, void (*fn)(cmap_spla
 }
 
 #ifdef CHECK_SPLAY
+
+static int
+tree_has_overlap(cmap_splay *tree, int node, int low, int high)
+{
+	if (tree[node].left != EMPTY)
+		if (tree_has_overlap(tree, tree[node].left, low, high))
+			return 1;
+	if (tree[node].right != EMPTY)
+		if (tree_has_overlap(tree, tree[node].right, low, high))
+			return 1;
+	return (tree[node].low < low && low < tree[node].high) || (tree[node].low < high && high < tree[node].high);
+}
+
 static void
 do_check(cmap_splay *node, void *arg)
 {
@@ -451,6 +486,7 @@ do_check(cmap_splay *node, void *arg)
 		tree[node->left].high < node->low));
 	assert(node->right == EMPTY || (tree[node->right].parent == num &&
 		node->high < tree[node->right].low));
+	assert(!tree_has_overlap(tree, num, node->low, node->high));
 }
 
 static void
@@ -471,6 +507,9 @@ add_range(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned int high,
 {
 	int current;
 	cmap_splay *tree;
+	int i;
+	int inrange = 0;
+	unsigned int k, count;
 
 	if (low > high)
 	{
@@ -478,6 +517,28 @@ add_range(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned int high,
 		return;
 	}
 
+	if (cmap->codespace_len == 0)
+	{
+		fz_warn(ctx, "CMap is missing codespace range");
+		pdf_add_codespace(ctx, cmap, 0, 65535, 2);
+	}
+
+	count = high - low + 1;
+	for (k = 0; k < count; k++) {
+		unsigned int c = low + k;
+
+		inrange = 0;
+		for (i = 0; i < cmap->codespace_len; i++) {
+			if (cmap->codespace[i].low <= c && c <= cmap->codespace[i].high)
+				inrange = 1;
+		}
+		if (!inrange)
+		{
+			fz_warn(ctx, "ignoring CMap range (%u-%u) that is outside of the codespace", low, high);
+			return;
+		}
+	}
+
 	tree = cmap->tree;
 
 	if (cmap->tlen)
@@ -508,6 +569,12 @@ add_range(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned int high,
 					tree[current].low = high + 1;
 					if (tree[current].low > tree[current].high)
 					{
+						/* update lt/gt references that will be moved/stale after deleting current */
+						if (gt == (unsigned int) cmap->tlen - 1)
+							gt = current;
+						if (lt == (unsigned int) cmap->tlen - 1)
+							lt = current;
+						/* delete_node() moves the element at cmap->tlen-1 into current */
 						move = delete_node(cmap, current);
 						current = EMPTY;
 						continue;
@@ -588,7 +655,7 @@ add_range(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned int high,
 	if (cmap->tlen == cmap->tcap)
 	{
 		int new_cap = cmap->tcap ? cmap->tcap * 2 : 256;
-		tree = cmap->tree = fz_resize_array(ctx, cmap->tree, new_cap, sizeof *cmap->tree);
+		tree = cmap->tree = fz_realloc_array(ctx, cmap->tree, new_cap, cmap_splay);
 		cmap->tcap = new_cap;
 	}
 	tree[cmap->tlen].low = low;
@@ -631,7 +698,7 @@ add_mrange(fz_context *ctx, pdf_cmap *cmap, unsigned int low, int *out, int len)
 	if (cmap->dlen + len + 1 > cmap->dcap)
 	{
 		int new_cap = cmap->dcap ? cmap->dcap * 2 : 256;
-		cmap->dict = fz_resize_array(ctx, cmap->dict, new_cap, sizeof *cmap->dict);
+		cmap->dict = fz_realloc_array(ctx, cmap->dict, new_cap, int);
 		cmap->dcap = new_cap;
 	}
 	out_pos = cmap->dlen;
@@ -642,45 +709,52 @@ add_mrange(fz_context *ctx, pdf_cmap *cmap, unsigned int low, int *out, int len)
 	add_range(ctx, cmap, low, low, out_pos, 1, 1);
 }
 
-/*
- * Add a range of contiguous one-to-one mappings (ie 1..5 maps to 21..25)
- */
 void
 pdf_map_range_to_range(fz_context *ctx, pdf_cmap *cmap, unsigned int low, unsigned int high, int out)
 {
 	add_range(ctx, cmap, low, high, out, 1, 0);
 }
 
-/*
- * Add a single one-to-many mapping.
- */
 void
-pdf_map_one_to_many(fz_context *ctx, pdf_cmap *cmap, unsigned int low, int *values, int len)
+pdf_map_one_to_many(fz_context *ctx, pdf_cmap *cmap, unsigned int low, int *values, size_t len)
 {
+	/* Decode unicode surrogate pairs. */
+	/* Only the *-UCS2 CMaps use one-to-many mappings, so assuming unicode should be safe. */
+	if (len >= 2)
+	{
+		size_t i, j;
+		/* Look for mranges with either multiple surrogate pairs in, or surrogate pairs
+		 * with other chars. See bug 706131. */
+		for (i = 0, j = 0; i < len; i++, j++)
+		{
+			int hi = values[i];
+			if (hi >= 0xd800 && hi < 0xdc00 && i < len-1)
+			{
+				int lo = values[i+1];
+				if (lo >= 0xdc00 && lo < 0xe000)
+				{
+					hi = ((hi - 0xD800) << 10) + (lo - 0xDC00) + 0x10000;
+					i++;
+				}
+			}
+			values[j] = hi;
+		}
+		len = j;
+	}
+
 	if (len == 1)
 	{
 		add_range(ctx, cmap, low, low, values[0], 1, 0);
 		return;
 	}
 
-	/* Decode unicode surrogate pairs. */
-	/* Only the *-UCS2 CMaps use one-to-many mappings, so assuming unicode should be safe. */
-	if (len == 2 &&
-		values[0] >= 0xD800 && values[0] <= 0xDBFF &&
-		values[1] >= 0xDC00 && values[1] <= 0xDFFF)
-	{
-		int rune = ((values[0] - 0xD800) << 10) + (values[1] - 0xDC00) + 0x10000;
-		add_range(ctx, cmap, low, low, rune, 1, 0);
-		return;
-	}
-
 	if (len > PDF_MRANGE_CAP)
 	{
 		fz_warn(ctx, "ignoring one to many mapping in cmap %s", cmap->cmap_name);
 		return;
 	}
 
-	add_mrange(ctx, cmap, low, values, len);
+	add_mrange(ctx, cmap, low, values, (int)len);
 }
 
 static void
@@ -737,11 +811,11 @@ pdf_sort_cmap(fz_context *ctx, pdf_cmap *cmap)
 	counts[2] = 0;
 	walk_splay(cmap->tree, cmap->ttop, count_node_types, &counts);
 
-	cmap->ranges = fz_malloc_array(ctx, counts[0], sizeof(*cmap->ranges));
+	cmap->ranges = Memento_label(fz_malloc_array(ctx, counts[0], pdf_range), "cmap_range");
 	cmap->rcap = counts[0];
-	cmap->xranges = fz_malloc_array(ctx, counts[1], sizeof(*cmap->xranges));
+	cmap->xranges = Memento_label(fz_malloc_array(ctx, counts[1], pdf_xrange), "cmap_xrange");
 	cmap->xcap = counts[1];
-	cmap->mranges = fz_malloc_array(ctx, counts[2], sizeof(*cmap->mranges));
+	cmap->mranges = Memento_label(fz_malloc_array(ctx, counts[2], pdf_mrange), "cmap_mrange");
 	cmap->mcap = counts[2];
 
 	walk_splay(cmap->tree, cmap->ttop, copy_node_types, cmap);
@@ -750,9 +824,6 @@ pdf_sort_cmap(fz_context *ctx, pdf_cmap *cmap)
 	cmap->tree = NULL;
 }
 
-/*
- * Lookup the mapping of a codepoint.
- */
 int
 pdf_lookup_cmap(pdf_cmap *cmap, unsigned int cpt)
 {
@@ -858,10 +929,6 @@ pdf_lookup_cmap_full(pdf_cmap *cmap, unsigned int cpt, int *out)
 	return 0;
 }
 
-/*
- * Use the codespace ranges to extract a codepoint from a
- * multi-byte encoded string.
- */
 int
 pdf_decode_cmap(pdf_cmap *cmap, unsigned char *buf, unsigned char *end, unsigned int *cpt)
 {
@@ -892,3 +959,19 @@ pdf_decode_cmap(pdf_cmap *cmap, unsigned char *buf, unsigned char *end, unsigned
 	*cpt = 0;
 	return 1;
 }
+
+size_t
+pdf_cmap_size(fz_context *ctx, pdf_cmap *cmap)
+{
+	if (cmap == NULL)
+		return 0;
+	if (cmap->storable.refs < 0)
+		return 0;
+
+	return pdf_cmap_size(ctx, cmap->usecmap) +
+		cmap->rcap * sizeof *cmap->ranges +
+		cmap->xcap * sizeof *cmap->xranges +
+		cmap->mcap * sizeof *cmap->mranges +
+		cmap->tcap * sizeof *cmap->tree +
+		sizeof(*cmap);
+}
