@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
@@ -74,6 +74,9 @@ struct pdf_gstate
 	pdf_material stroke;
 	pdf_material fill;
 
+	/* pattern paint type 2 */
+	int ismask;
+
 	/* text state */
 	pdf_text_state text;
 
@@ -463,6 +466,8 @@ pdf_show_pattern(fz_context *ctx, pdf_run_processor *pr, pdf_pattern *pat, int p
 
 	if (pat->ismask)
 	{
+		/* Inhibit any changes to the color since we're drawing an uncolored pattern. */
+		gstate->ismask = 1;
 		pdf_unset_pattern(ctx, pr, PDF_FILL);
 		pdf_unset_pattern(ctx, pr, PDF_STROKE);
 		if (what == PDF_FILL)
@@ -1213,6 +1218,8 @@ pdf_init_gstate(fz_context *ctx, pdf_gstate *gs, fz_matrix ctm)
 
 	gs->fill.color_params = fz_default_color_params;
 	gs->stroke.color_params = fz_default_color_params;
+
+	gs->ismask = 0;
 }
 
 static void
@@ -1236,6 +1243,10 @@ pdf_set_colorspace(fz_context *ctx, pdf_run_processor *pr, int what, fz_colorspa
 
 	gstate = pdf_flush_text(ctx, pr);
 
+	/* Don't change color if we're drawing an uncolored pattern tile! */
+	if (gstate->ismask)
+		return;
+
 	mat = what == PDF_FILL ? &gstate->fill : &gstate->stroke;
 
 	fz_drop_colorspace(ctx, mat->colorspace);
@@ -1264,6 +1275,10 @@ pdf_set_color(fz_context *ctx, pdf_run_processor *pr, int what, float *v)
 
 	gstate = pdf_flush_text(ctx, pr);
 
+	/* Don't change color if we're drawing an uncolored pattern tile! */
+	if (gstate->ismask)
+		return;
+
 	mat = what == PDF_FILL ? &gstate->fill : &gstate->stroke;
 
 	switch (mat->kind)
@@ -1516,8 +1531,14 @@ begin_oc(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val, pdf_cycle_list
 	obj = pdf_dict_get(ctx, val, PDF_NAME(Name));
 	if (obj)
 	{
+		const char *name = "";
 		pdf_flush_text(ctx, proc);
-		push_begin_layer(ctx, proc, pdf_to_name(ctx, obj));
+		if (pdf_is_name(ctx, obj))
+			name = pdf_to_name(ctx, obj);
+		else if (pdf_is_string(ctx, obj))
+			name = pdf_to_text_string(ctx, obj);
+
+		push_begin_layer(ctx, proc, name);
 		return;
 	}
 
@@ -1606,24 +1627,67 @@ pop_structure_to(fz_context *ctx, pdf_run_processor *proc, pdf_obj *common)
 	}
 }
 
+struct line
+{
+	pdf_obj *obj;
+	struct line *child;
+};
+
+static pdf_obj *
+find_most_recent_common_ancestor_imp(fz_context *ctx, pdf_obj *a, struct line *line_a, pdf_obj *b, struct line *line_b, pdf_cycle_list *cycle_up_a, pdf_cycle_list *cycle_up_b)
+{
+	struct line line;
+	pdf_obj *common = NULL;
+	pdf_cycle_list cycle;
+	pdf_obj *parent;
+
+	/* First ascend one lineage. */
+	if (pdf_is_dict(ctx, a))
+	{
+		if (pdf_cycle(ctx, &cycle, cycle_up_a, a))
+			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in structure tree");
+		line.obj = a;
+		line.child = line_a;
+		parent = pdf_dict_get(ctx, a, PDF_NAME(P));
+		return find_most_recent_common_ancestor_imp(ctx, parent, &line, b, NULL, &cycle, NULL);
+	}
+	/* Then ascend the other lineage. */
+	else if (pdf_is_dict(ctx, b))
+	{
+		if (pdf_cycle(ctx, &cycle, cycle_up_b, b))
+			fz_throw(ctx, FZ_ERROR_GENERIC, "cycle in structure tree");
+		line.obj = b;
+		line.child = line_b;
+		parent = pdf_dict_get(ctx, b, PDF_NAME(P));
+		return find_most_recent_common_ancestor_imp(ctx, a, line_a, parent, &line, cycle_up_a, &cycle);
+	}
+
+	/* Once both lineages are know, traverse top-down to find most recent common ancestor. */
+	while (line_a && line_b && !pdf_objcmp(ctx, line_a->obj, line_b->obj))
+	{
+		common = line_a->obj;
+		line_a = line_a->child;
+		line_b = line_b->child;
+	}
+	return common;
+}
+
+static pdf_obj *
+find_most_recent_ancestor(fz_context *ctx, pdf_obj *a, pdf_obj *b)
+{
+	if (!pdf_is_dict(ctx, a) || !pdf_is_dict(ctx, b))
+		return NULL;
+	return find_most_recent_common_ancestor_imp(ctx, a, NULL, b, NULL, NULL, NULL);
+}
+
 static void
 send_begin_structure(fz_context *ctx, pdf_run_processor *proc, pdf_obj *mc_dict)
 {
-	pdf_obj *common;
-	pdf_obj *parent_tree_root = pdf_dict_getl(ctx, pdf_trailer(ctx, proc->doc), PDF_NAME(Root), PDF_NAME(StructTreeRoot), PDF_NAME(ParentTree), NULL);
+	pdf_obj *common = NULL;
 
 	/* We are currently nested in A,B,C,...E,F,mcid_sent. We want to update to
-	 * being in A,B,C,...G,H,mc_dict. So we need to find the lowest common
-	 * point. We know that the structure is a tree, so no cycles to worry about.
-	 * Live with an n^2 algorithm for now. */
-	for (common = mc_dict; common != NULL && pdf_objcmp(ctx, common, parent_tree_root); common = pdf_dict_get(ctx, common, PDF_NAME(P)))
-	{
-		pdf_obj *o;
-
-		for (o = proc->mcid_sent; o != NULL && pdf_objcmp(ctx, o, common) && pdf_objcmp(ctx, o, parent_tree_root); o = pdf_dict_get(ctx, o, PDF_NAME(P)));
-		if (!pdf_objcmp(ctx, o, common))
-			break;
-	}
+	 * being in A,B,C,...G,H,mc_dict. So we need to find the lowest common point. */
+	common = find_most_recent_ancestor(ctx, proc->mcid_sent, mc_dict);
 
 	/* So, we need to pop everything up to common (i.e. everything below common will be closed). */
 	pop_structure_to(ctx, proc, common);
@@ -1665,6 +1729,9 @@ push_marked_content(fz_context *ctx, pdf_run_processor *proc, const char *tagstr
 	fz_structure standard;
 	pdf_obj *mc_dict = NULL;
 
+	/* Flush any pending text so it's not in the wrong layer. */
+	pdf_flush_text(ctx, proc);
+
 	if (!tagstr)
 		tagstr = "Untitled";
 	tag = pdf_new_name(ctx, tagstr);
@@ -1752,6 +1819,9 @@ pop_marked_content(fz_context *ctx, pdf_run_processor *proc, int neat)
 		return;
 	}
 
+	/* Make sure that any pending text is written into the correct layer. */
+	pdf_flush_text(ctx, proc);
+
 	/* Close structure/layers here, in reverse order to how we opened them. */
 	fz_try(ctx)
 	{
